#ifndef _ASMARM_STACK_H_
#define _ASMARM_STACK_H_

#ifndef _STACK_H_
#error Do not directly include <asm/stack.h>. Just use <stack.h>.
#endif

#define HAVE_ARCH_BACKTRACE_FRAME
#define HAVE_ARCH_BACKTRACE

#endif
